#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define TEMP_THRESHOLD 80000
#define CLOCK_STEP 50
#define MIN_CLOCK 800
#define MAX_CLOCK 1500
#define INTERVAL 5
#define K_FACTOR 1e-18
#define R_THERMAL 0.5
#define AMB_TEMP 25000
#define THETA_MAX 85000
#define BETA 0.1

int read_gpu_temp() {
    FILE *fp = popen("nvidia-smi --query-gpu=temperature.gpu --format=csv,noheader", "r");
    if (!fp) return -1;
    char buffer[16];
    fgets(buffer, sizeof(buffer), fp);
    pclose(fp);
    return atoi(buffer) * 1000;
}

int get_gpu_clock() {
    FILE *fp = popen("nvidia-smi --query-gpu=clocks.gr --format=csv,noheader", "r");
    if (!fp) return -1;
    char buffer[16];
    fgets(buffer, sizeof(buffer), fp);
    pclose(fp);
    return atoi(buffer);
}

void set_gpu_clock(int clock) {
    char cmd[64];
    snprintf(cmd, sizeof(cmd), "nvidia-smi -lgc %d", clock);
    system(cmd);
}

void predict_performance(int clock, double load, double workload, double *power, double *time, double *temp) {
    double freq = clock * 1e6;
    *power = K_FACTOR * freq * freq * load;
    *time = workload / (freq * 1000);
    *temp = AMB_TEMP + *power * R_THERMAL * 1000;
}

int should_increase_clock(double current_time, double new_time, double current_temp, double new_temp) {
    double delta_time = current_time - new_time;
    double delta_temp = new_temp - current_temp;
    double time_ratio = delta_time / current_time;
    double temp_ratio = delta_temp / (THETA_MAX - current_temp);
    return time_ratio > BETA * temp_ratio;
}

int main() {
    int current_clock = get_gpu_clock();
    if (current_clock < 0) return 1;

    double workload = 1e12;
    double load = 0.9;
    double current_power, current_time, current_pred_temp;
    double new_power, new_time, new_pred_temp;

    while (1) {
        int temp = read_gpu_temp();
        if (temp < 0) break;

        predict_performance(current_clock, load, workload, &current_power, &current_time, &current_pred_temp);

        printf("Temp: %.1f°C, Clock: %d MHz, Pred Power: %.3f W, Pred Time: %.3f s, Pred Temp: %.1f°C\n",
               temp / 1000.0, current_clock, current_power, current_time, current_pred_temp / 1000.0);

        if (temp > TEMP_THRESHOLD && current_clock > MIN_CLOCK) {
            int new_clock = current_clock - CLOCK_STEP;
            if (new_clock < MIN_CLOCK) new_clock = MIN_CLOCK;
            set_gpu_clock(new_clock);
            current_clock = new_clock;
        } else if (temp < (TEMP_THRESHOLD - 10000) && current_clock < MAX_CLOCK) {
            int new_clock = current_clock + CLOCK_STEP;
            if (new_clock > MAX_CLOCK) new_clock = MAX_CLOCK;
            predict_performance(new_clock, load, workload, &new_power, &new_time, &new_pred_temp);
            if (should_increase_clock(current_time, new_time, temp, new_pred_temp)) {
                set_gpu_clock(new_clock);
                current_clock = new_clock;
            }
        }

        sleep(INTERVAL);
    }

    return 0;
}